------------------------------------------------------------------------
-- Data type used to distinguish between structural and βη equalities
------------------------------------------------------------------------

-- # import Eq

infixl 50 ∨

data EqKind : Set where
  structural : EqKind
  βη         : EqKind

(∨) : EqKind -> EqKind -> EqKind
βη         ∨ k = βη
structural ∨ k = k

abstract

  structuralRightUnit : (k : EqKind) -> k ∨ structural ≡ k
  structuralRightUnit structural = refl
  structuralRightUnit βη         = refl

  βηRightZero : (k : EqKind) -> k ∨ βη ≡ βη
  βηRightZero structural = refl
  βηRightZero βη         = refl

  idempotent∨ : (k : EqKind) -> k ∨ k ≡ k
  idempotent∨ structural = refl
  idempotent∨ βη         = refl

  associative∨
    : (k₁, k₂, k₃ : EqKind) -> k₁ ∨ (k₂ ∨ k₃) ≡ (k₁ ∨ k₂) ∨ k₃
  associative∨ structural _ _ = refl
  associative∨ βη         _ _ = refl

  commutative∨ : (k₁, k₂ : EqKind) -> k₁ ∨ k₂ ≡ k₂ ∨ k₁
  commutative∨ structural k₂ = sym $ structuralRightUnit k₂
  commutative∨ βη         k₂ = sym $ βηRightZero k₂
