------------------------------------------------------------------------
-- Defined casts, and some lemmas depending on these casts
------------------------------------------------------------------------

-- # import Explicit.Types
-- # import Explicit.Equivalence
-- # import Explicit.SubstitutionLemmas
-- # import Explicit.IndicesEqual
-- # import Explicit.Misc
-- # import Explicit.MoreSubstitutions

infixl 50 ◅ʳ, ++ʳ, ++ʳCong
infixl 45 ∷⊢, ∷∋, ∷✶, ∷⇒ˡ, ∷⇒ʳ, ∷→

------------------------------------------------------------------------
-- Substitution casts

mutual

  (∷⇒ˡ) (Γ¹, Γ², Δ : Ctxt)
    : Subst Γ¹ Δ -> Γ¹ ≜ Γ² -> Subst Γ² Δ
  (∷⇒ˡ) {_}      {_}      {Δ} ρ εCong           = ∅ Δ
  (∷⇒ˡ) {_ ▻ τ¹} {_ ▻ τ²}     ρ (eqΓ ▻Cong eqτ) =
    (init ρ ∷⇒ˡ eqΓ) ▷ (last ρ ∷≡⊢ lemma)
    where
    abstract
      lemma : τ¹ / init ρ ≛ τ² / (init ρ ∷⇒ˡ eqΓ)
      lemma = substTyCong eqτ (castSubstˡEq (init ρ) eqΓ)

  castSubstˡEq (Γ¹, Γ², Δ : Ctxt)
    : (ρ : Subst Γ¹ Δ) -> (eqΓ : Γ¹ ≜ Γ²) -> ρ ≘ ρ ∷⇒ˡ eqΓ
  castSubstˡEq = helper
    where
    abstract
      helper (Γ¹, Γ², Δ : Ctxt)
        : (ρ : Subst Γ¹ Δ) -> (eqΓ : Γ¹ ≜ Γ²) -> ρ ≘ ρ ∷⇒ˡ eqΓ
      helper ρ εCong           = emptyCtxtSubstIs∅ ρ
      helper ρ (eqΓ ▻Cong eqτ) =
        nonEmptyCtxtSubstIs▷ ρ
          `transSubst`
        Cong▷ (castSubstˡEq (init ρ) eqΓ)
              eqτ
              (symTm $ castTmEq (last ρ) _)

abstract

  generalisedId (Γ, Δ : Ctxt)
    : Γ ≜ Δ -> Subst Γ Δ
  generalisedId {Γ} {Δ} eqΓ = id Δ ∷⇒ˡ symCtxt eqΓ

  generalisedIdIsId (Γ, Δ : Ctxt)
    : (eqΓ : Γ ≜ Δ) -> generalisedId eqΓ ≘ id Γ
  generalisedIdIsId {Γ} {Δ} eqΓ =
    symSubst (castSubstˡEq (id Δ) (symCtxt eqΓ))
      `transSubst`
    idCong (symCtxt eqΓ)

  generalisedIdVanishesTy (Γ, Δ : Ctxt)
    : (τ : Ty Γ) -> (eqΓ : Γ ≜ Δ) -> τ / generalisedId eqΓ ≛ τ
  generalisedIdVanishesTy τ eqΓ =
    substTyCong (reflTy τ) (generalisedIdIsId eqΓ)
      `transTy`
    idVanishesTy τ

  generalisedIdVanishesTm (Γ, Δ : Ctxt) (τ : Ty Γ)
    : (t : Γ ⊢ τ) -> (eqΓ : Γ ≜ Δ) -> t /⊢ generalisedId eqΓ ≈⊢ t
  generalisedIdVanishesTm t eqΓ =
    substTmCong (reflTm t) (generalisedIdIsId eqΓ)
      `transTm`
    idVanishesTm t

  generalisedIdCong (Γ¹, Δ¹ : Ctxt)
                    (Γ², Δ² : Ctxt)
    :  (eqΓ¹ : Γ¹ ≜ Δ¹) -> (eqΓ² : Γ² ≜ Δ²) -> Γ¹ ≜ Γ²
    -> generalisedId eqΓ¹ ≘ generalisedId eqΓ²
  generalisedIdCong eqΓ¹ eqΓ² eqΓ =
    betweenSubst (generalisedIdIsId eqΓ¹)
                 (generalisedIdIsId eqΓ²) $
      idCong eqΓ

  (∷⇒ʳ) (Γ, Δ¹, Δ² : Ctxt)
    : Subst Γ Δ¹ -> Δ¹ ≜ Δ² -> Subst Γ Δ²
  ρ ∷⇒ʳ eqΔ = ρ ⊙ generalisedId eqΔ

  castSubstʳEq (Γ, Δ¹, Δ² : Ctxt)
    : (ρ : Subst Γ Δ¹) -> (eqΔ : Δ¹ ≜ Δ²) -> ρ ≘ ρ ∷⇒ʳ eqΔ
  castSubstʳEq {Γ} {Δ¹} ρ eqΔ =
    symSubst (idRightUnit ρ)
      `transSubst`
    compCong (reflSubst ρ) (symSubst $ generalisedIdIsId eqΔ)

------------------------------------------------------------------------
-- Type casts

(∷✶) (Γ, Δ : Ctxt)
  : Ty Γ -> (eqΓ : Γ ≜ Δ) -> Ty Δ
τ ∷✶ eqΓ = τ / generalisedId eqΓ

abstract

  castTyEq (Γ, Δ : Ctxt)
    : (τ : Ty Γ) -> (eqΓ : Γ ≜ Δ) -> τ ≛ τ ∷✶ eqΓ
  castTyEq τ eqΓ =
    symTy (idVanishesTy τ)
      `transTy`
    substTyCong (reflTy τ) (symSubst $ generalisedIdIsId eqΓ)

  castTyEqˡ (Δ : Ctxt)
            (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
            (Γ² : Ctxt) (τ² : Ty Γ²)
    : (eqΓ : Γ¹ ≜ Δ) -> τ¹ ≛ τ² -> τ¹ ∷✶ eqΓ ≛ τ²
  castTyEqˡ eqΓ eqτ = symTy (castTyEq _ eqΓ) `transTy` eqτ

  castTyEqʳ (Δ : Ctxt)
            (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
            (Γ² : Ctxt) (τ² : Ty Γ²)
    : (eqΓ : Γ² ≜ Δ) -> τ¹ ≛ τ² -> τ¹ ≛ τ² ∷✶ eqΓ
  castTyEqʳ eqΓ eqτ = eqτ `transTy` castTyEq _ eqΓ

  castTyCong (Γ¹, Γ², Δ¹, Δ² : Ctxt) (τ¹ : Ty Γ¹) (τ² : Ty Γ²)
    :  (eqΓ¹ : Γ¹ ≜ Δ¹) -> (eqΓ² : Γ² ≜ Δ²)
    -> τ¹ ≛ τ² -> τ¹ ∷✶ eqΓ¹ ≛ τ² ∷✶ eqΓ²
  castTyCong eqΓ¹ eqΓ² eqτ = castTyEqˡ eqΓ¹ $ castTyEqʳ eqΓ² eqτ

------------------------------------------------------------------------
-- Term casts

abstract

  (∷⊢) (Γ¹, Γ² : Ctxt) (τ¹ : Ty Γ¹) (τ² : Ty Γ²)
    : Γ¹ ⊢ τ¹ -> (eqτ : τ¹ ≛ τ²) -> Γ² ⊢ τ²
  (∷⊢) {Γ¹} {Γ²} {τ¹} {τ²} t eqτ =
    t /⊢ generalisedId (tyEqCtxtEq eqτ) ∷≡⊢ lemma
    where
    lemma : τ¹ ∷✶ tyEqCtxtEq eqτ ≛ τ²
    lemma =
      symTy (castTyEq τ¹ (tyEqCtxtEq eqτ))
        `transTy`
      eqτ

  generalCastTmEq (Γ¹, Γ² : Ctxt) (τ¹ : Ty Γ¹) (τ² : Ty Γ²)
    : (t : Γ¹ ⊢ τ¹) -> (eqτ : τ¹ ≛ τ²) -> t ≈⊢ t ∷⊢ eqτ
  generalCastTmEq t eqτ = castTmEqʳ _ $
    symTm (idVanishesTm t)
      `transTm`
    substTmCong (reflTm t) (symSubst $ generalisedIdIsId eqΓ)
    where eqΓ = tyEqCtxtEq eqτ

  generalCastTmCong (Γ₁¹ : Ctxt) (τ₁¹ : Ty Γ₁¹) (t¹ : Γ₁¹ ⊢ τ₁¹)
                    (Γ₂¹ : Ctxt) (τ₂¹ : Ty Γ₂¹)
                    (Γ₁² : Ctxt) (τ₁² : Ty Γ₁²) (t² : Γ₁² ⊢ τ₁²)
                    (Γ₂² : Ctxt) (τ₂² : Ty Γ₂²)
    :  (eqτ¹ : τ₁¹ ≛ τ₂¹) -> (eqτ² : τ₁² ≛ τ₂²)
    -> t¹ ≈⊢ t² -> t¹ ∷⊢ eqτ¹ ≈⊢ t² ∷⊢ eqτ²
  generalCastTmCong eqτ¹ eqτ² eqt =
    betweenTm (symTm $ generalCastTmEq _ eqτ¹)
              (symTm $ generalCastTmEq _ eqτ²)
              eqt

------------------------------------------------------------------------
-- Variable casts

-- These casts are trickier to define, since there is no substitution
-- application operation (like /⊢) taking variables to variables.

abstract

  castVarCtxt (Γ, Δ : Ctxt) (τ : Ty Γ)
    : (v : Γ ∋ τ) -> (eqΓ : Γ ≜ Δ) -> Δ ∋ τ ∷✶ eqΓ
  castVarCtxt {Γ ▻ σ¹} {Δ ▻ σ²} (vz σ¹) (eqΓ ▻Cong eqσ) =
    vz σ² ∷≡∋ lemma
    where
    lemma : σ² / wk σ² ≛ σ¹ / wk σ¹ ∷✶ eqΓ ▻Cong eqσ
    lemma =
      symTy (substTyCong eqσ (wkCong eqσ))
        `transTy`
      castTyEq _ _
  castVarCtxt {Γ ▻ σ¹} {Δ ▻ σ²} (vs {Γ} {τ¹} v σ¹) (eqΓ ▻Cong eqσ) =
    vs (castVarCtxt v eqΓ) σ² ∷≡∋ lemma
    where
    lemma : (τ¹ ∷✶ eqΓ) / wk σ² ≛ τ¹ / wk σ¹ ∷✶ eqΓ ▻Cong eqσ
    lemma =
      symTy (substTyCong (castTyEq _ _) (wkCong eqσ))
        `transTy`
      castTyEq _ _
  castVarCtxt (v ∷≡∋ eqτ) eqΓ = castVarCtxt v eqΓ ∷≡∋ castTyCong _ _ eqτ

  castVarCtxtEq (Γ, Δ : Ctxt) (τ : Ty Γ)
    : (v : Γ ∋ τ) -> (eqΓ : Γ ≜ Δ) -> v ≈∋ castVarCtxt v eqΓ
  castVarCtxtEq (vz σ¹)   (eqΓ ▻Cong eqσ) = castVarEqʳ _ $ vzCong eqσ
  castVarCtxtEq (vs v σ¹) (eqΓ ▻Cong eqσ) = castVarEqʳ _ $
    vsCong (castVarCtxtEq v eqΓ) eqσ
  castVarCtxtEq (v ∷≡∋ eqτ) eqΓ = castVarCong _ _ $ castVarCtxtEq v eqΓ

  (∷∋) (Γ¹, Γ² : Ctxt) (τ¹ : Ty Γ¹) (τ² : Ty Γ²)
    : Γ¹ ∋ τ¹ -> (eqτ : τ¹ ≛ τ²) -> Γ² ∋ τ²
  (∷∋) {Γ¹} {Γ²} {τ¹} {τ²} v eqτ =
    castVarCtxt v (tyEqCtxtEq eqτ) ∷≡∋ lemma
    where
    lemma : τ¹ ∷✶ tyEqCtxtEq eqτ ≛ τ²
    lemma =
      symTy (castTyEq τ¹ (tyEqCtxtEq eqτ))
        `transTy`
      eqτ

  generalCastVarEq (Γ¹, Γ² : Ctxt) (τ¹ : Ty Γ¹) (τ² : Ty Γ²)
    : (v : Γ¹ ∋ τ¹) -> (eqτ : τ¹ ≛ τ²) -> v ≈∋ v ∷∋ eqτ
  generalCastVarEq {Γ¹} {Γ²} {τ¹} {τ²} v eqτ =
    castVarEqʳ (symTy (castTyEq τ¹ (tyEqCtxtEq eqτ))
                  `transTy`
                eqτ)
               (castVarCtxtEq v (tyEqCtxtEq eqτ))

------------------------------------------------------------------------
-- RCtxt casts

abstract

  (∷→) (Γ, Δ : Ctxt) : RCtxt Γ -> Γ ≜ Δ -> RCtxt Δ
  Γ' ∷→ eqΓ = Γ' /→ generalisedId eqΓ

  castRCtxtEq (Γ, Δ : Ctxt)
    : (Γ' : RCtxt Γ) -> (eqΓ : Γ ≜ Δ) -> Γ' ≜→ Γ' ∷→ eqΓ
  castRCtxtEq Γ' eqΓ =
    symRCtxt (idVanishesRCtxt Γ')
      `transRCtxt`
    substRCtxtCong (reflRCtxt Γ') (symSubst $ generalisedIdIsId eqΓ)

  castRCtxtCong (Γ¹, Δ¹ : Ctxt) (Γ'¹ : RCtxt Γ¹)
                (Γ², Δ² : Ctxt) (Γ'² : RCtxt Γ²)
    :  (eqΓ¹ : Γ¹ ≜ Δ¹) -> (eqΓ² : Γ² ≜ Δ²)
    -> Γ'¹ ≜→ Γ'² -> Γ'¹ ∷→ eqΓ¹ ≜→ Γ'² ∷→ eqΓ²
  castRCtxtCong _ _ eqΓ' =
    betweenRCtxt (symRCtxt $ castRCtxtEq _ _)
                 (symRCtxt $ castRCtxtEq _ _) $
      eqΓ'

------------------------------------------------------------------------
-- Prepending to right contexts

mutual

  (◅ʳ) (Γ : Ctxt) : (τ : Ty Γ) -> RCtxt (Γ ▻ τ) -> RCtxt Γ
  τ ◅ʳ εʳ (Γ ▻ τ) = singletonʳ τ
  τ ◅ʳ (Γ' ▻ʳ σ)  = (τ ◅ʳ Γ') ▻ʳ (σ ∷✶ snoc++Is++Cons Γ')

  snoc++Is++Cons (Γ : Ctxt) (τ : Ty Γ)
    : (Γ' : RCtxt (Γ ▻ τ)) -> Γ ▻ τ ++ Γ' ≜ Γ ++ (τ ◅ʳ Γ')
  snoc++Is++Cons = helper
    where
    abstract
      helper (Γ : Ctxt) (τ : Ty Γ)
        : (Γ' : RCtxt (Γ ▻ τ)) -> Γ ▻ τ ++ Γ' ≜ Γ ++ (τ ◅ʳ Γ')
      helper (εʳ (Γ ▻ τ)) = reflCtxt _
      helper (Γ' ▻ʳ σ)    = lemma ▻Cong castTyEq σ lemma
        where lemma = snoc++Is++Cons Γ'

abstract

  Cong◅ʳ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (Γ'¹ : RCtxt (Γ¹ ▻ τ¹))
         (Γ² : Ctxt) (τ² : Ty Γ²) (Γ'² : RCtxt (Γ² ▻ τ²))
    : Γ'¹ ≜→ Γ'² -> τ¹ ◅ʳ Γ'¹ ≜→ τ² ◅ʳ Γ'²
  Cong◅ʳ (εʳCong (eqΓ ▻Cong eqτ)) = singletonʳCong eqτ
  Cong◅ʳ (eqΓ' ▻ʳCong eqσ)        =
    Cong◅ʳ eqΓ' ▻ʳCong castTyCong _ _ eqσ

------------------------------------------------------------------------
-- Appending to right contexts

mutual

  (++ʳ) (Γ : Ctxt) : (Γ' : RCtxt Γ) -> RCtxt (Γ ++ Γ') -> RCtxt Γ
  Γ'₁ ++ʳ εʳ _       = Γ'₁
  Γ'₁ ++ʳ (Γ'₂ ▻ʳ τ) = (Γ'₁ ++ʳ Γ'₂) ▻ʳ (τ ∷✶ assoc++ Γ'₁ Γ'₂)

  assoc++ (Γ : Ctxt)
    :  (Γ'₁ : RCtxt Γ) -> (Γ'₂ : RCtxt (Γ ++ Γ'₁))
    -> (Γ ++ Γ'₁) ++ Γ'₂ ≜ Γ ++ (Γ'₁ ++ʳ Γ'₂)
  assoc++ _ = helper
    where
    abstract
      helper (Γ : Ctxt) (Γ'₁ : RCtxt Γ)
        :  (Γ'₂ : RCtxt (Γ ++ Γ'₁))
        -> (Γ ++ Γ'₁) ++ Γ'₂ ≜ Γ ++ (Γ'₁ ++ʳ Γ'₂)
      helper (εʳ _)     = reflCtxt _
      helper (Γ'₂ ▻ʳ τ) = lemma ▻Cong castTyEq τ lemma
        where lemma = assoc++ _ Γ'₂

abstract

  (++ʳCong) (Γ¹ : Ctxt) (Γ'₁¹ : RCtxt Γ¹) (Γ'₂¹ : RCtxt (Γ¹ ++ Γ'₁¹))
            (Γ² : Ctxt) (Γ'₁² : RCtxt Γ²) (Γ'₂² : RCtxt (Γ² ++ Γ'₁²))
    : Γ'₁¹ ≜→ Γ'₁² -> Γ'₂¹ ≜→ Γ'₂² -> Γ'₁¹ ++ʳ Γ'₂¹ ≜→ Γ'₁² ++ʳ Γ'₂²
  eqΓ'₁ ++ʳCong εʳCong _           = eqΓ'₁
  eqΓ'₁ ++ʳCong (eqΓ'₂ ▻ʳCong eqτ) =
    (eqΓ'₁ ++ʳCong eqΓ'₂) ▻ʳCong castTyCong _ _ eqτ

------------------------------------------------------------------------
-- Substitution lemmas

abstract

  wk⋆Splitsˡ (Γ : Ctxt) (σ : Ty Γ)
    : (Γ' : RCtxt (Γ ▻ σ)) -> wk⋆ (σ ◅ʳ Γ') ≘ wk σ ⊙ wk⋆ Γ'
  wk⋆Splitsˡ (εʳ (Γ ▻ σ)) =
    idLeftUnit _
      `transSubst`
    symSubst (idRightUnit _)
  wk⋆Splitsˡ (Γ' ▻ʳ τ) =
    compCong (wk⋆Splitsˡ Γ') (wkCong $ symTy $ castTyEq τ _)
      `transSubst`
    symSubst (compAssoc _ _ _)

  wk⋆SplitsˡTy (Γ : Ctxt) (σ : Ty Γ)
    :  (τ : Ty Γ) -> (Γ' : RCtxt (Γ ▻ σ))
    -> τ / wk⋆ (σ ◅ʳ Γ') ≛ τ / wk σ / wk⋆ Γ'
  wk⋆SplitsˡTy {Γ} {σ} τ Γ' =
    substTyCong (reflTy τ) (wk⋆Splitsˡ Γ')
      `transTy`
    compSplitsTy τ (wk σ) (wk⋆ Γ')
