------------------------------------------------------------------------
-- Extensional equality for substitutions
------------------------------------------------------------------------

-- It is often easier to work with the extensional equality /≘/,
-- introduced here, than with its first-order cousin ≘.

-- # import Explicit.Types
-- # import Explicit.Equivalence

infix 40 /≘/

------------------------------------------------------------------------
-- Equal substitutions have equal left and right contexts

-- More lemmas like these follow in IndicesEqual.

abstract

  substEqCtxtEqRight (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
                     (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
    : ρ¹ ≘ ρ² -> Δ¹ ≜ Δ²
  substEqCtxtEqRight (emptyCong eqΔ)        = eqΔ
  substEqCtxtEqRight (snocCong eqρ eqτ eqt) = substEqCtxtEqRight eqρ

  substEqCtxtEqLeft (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
                    (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
    : ρ¹ ≘ ρ² -> Γ¹ ≜ Γ²
  substEqCtxtEqLeft (emptyCong eqΔ)        = εCong
  substEqCtxtEqLeft (snocCong eqρ eqτ eqt) =
    substEqCtxtEqLeft eqρ ▻Cong eqτ

------------------------------------------------------------------------
-- Extensional substitution equality

-- If two transformations have the same behaviour on variables, and
-- the resulting contexts are equal, then the transformations are
-- extensionally equal.

ExtEq (Γ¹, Γ², Δ¹, Δ² : Ctxt) : Subst Γ¹ Δ¹ -> Subst Γ² Δ² -> Set
ExtEq {Γ¹} {Γ²} ρ¹ ρ² =
     (σ¹ : Ty Γ¹) -> (v¹ : Γ¹ ∋ σ¹)
  -> (σ² : Ty Γ²) -> (v² : Γ² ∋ σ²)
  -> v¹ ≈∋ v² -> var v¹ /⊢ ρ¹ ≈⊢ var v² /⊢ ρ²

data (/≘/) {Γ¹, Γ², Δ¹, Δ² : Ctxt}
           (ρ¹ : Subst Γ¹ Δ¹) (ρ² : Subst Γ² Δ²) : Set where
  extEq : Γ¹ ≜ Γ² -> Δ¹ ≜ Δ² -> ExtEq ρ¹ ρ² -> ρ¹ /≘/ ρ²

abstract

  extEqProof (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
             (σ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ σ¹)
             (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
             (σ² : Ty Γ²) (v² : Γ² ∋ σ²)
    : v¹ ≈∋ v² -> ρ¹ /≘/ ρ² -> var v¹ /⊢ ρ¹ ≈⊢ var v² /⊢ ρ²
  extEqProof eqv (extEq eqΓ eqΔ proof) = proof _ _ _ _ eqv

  initExtEqCong (Γ¹, Δ¹ : Ctxt) (τ¹ : Ty Γ¹) (ρ¹ : Subst (Γ¹ ▻ τ¹) Δ¹)
                (Γ², Δ² : Ctxt) (τ² : Ty Γ²) (ρ² : Subst (Γ² ▻ τ²) Δ²)
    : ρ¹ /≘/ ρ² -> init ρ¹ /≘/ init ρ²
  initExtEqCong (extEq (eqΓ₀ ▻Cong eqτ₀) eqΔ₀ proof₀) =
    extEq eqΓ₀ eqΔ₀ (proof eqτ₀ proof₀)
    where
    proof (Γ¹, Δ¹ : Ctxt) (τ¹ : Ty Γ¹) (ρ¹ : Subst (Γ¹ ▻ τ¹) Δ¹)
          (Γ², Δ² : Ctxt) (τ² : Ty Γ²) (ρ² : Subst (Γ² ▻ τ²) Δ²)
      : τ¹ ≛ τ² -> ExtEq ρ¹ ρ² -> ExtEq (wk τ¹ ⊙ ρ¹) (wk τ² ⊙ ρ²)
    proof {Γ¹} {Δ¹} {τ¹} {ρ¹} {Γ²} {Δ²} {τ²} {ρ²}
          eqτ proof' _ v¹ _ v² eqv =
      compSplitsTm (var v¹) (wk τ¹) ρ¹
        `transTm`
      substTmCong (substWk v¹ τ¹) (reflSubst ρ¹)
        `transTm`
      proof' _ (vs v¹ τ¹) _ (vs v² τ²) (vsCong eqv eqτ)
        `transTm`
      symTm (substTmCong (substWk v² τ²) (reflSubst ρ²))
        `transTm`
      symTm (compSplitsTm (var v²) (wk τ²) ρ²)

  extEqToSubstEq (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
                 (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
    : ρ¹ /≘/ ρ² -> ρ¹ ≘ ρ²
  extEqToSubstEq {ε} (extEq εCong eqΔ _) = emptyCong eqΔ
  extEqToSubstEq {Γ¹ ▻ σ¹} {Δ¹} {ρ¹} {Γ² ▻ σ²} {Δ²} {ρ²}
                 (extEq (eqΓ ▻Cong eqσ) eqΔ proof) =
    snocCong (extEqToSubstEq {Γ¹}
               (extEq eqΓ eqΔ (\_ _ _ _ eqv -> extEqProof eqv eqInitρ)))
             eqσ
             (proof _ (vz σ¹) _ (vz σ²) (vzCong eqσ))
    where eqInitρ = initExtEqCong (extEq (eqΓ ▻Cong eqσ) eqΔ proof)

  substEqToExtEq (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
                 (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
    : ρ¹ ≘ ρ² -> ρ¹ /≘/ ρ²
  substEqToExtEq eqρ =
    extEq (substEqCtxtEqLeft eqρ)
          (substEqCtxtEqRight eqρ)
          (\σ¹ v¹ σ² v² eqv -> substTmCong (varCong eqv) eqρ)
    -- Can probably be proved without substTmCong; i.e. substVarCong can
    -- probably be proved.

------------------------------------------------------------------------
-- Extensional substitution equality variants

-- A simpler variant which is useful when the starting contexts are
-- definitionally equal.

ExtEqˡ⁼ (Γ, Δ¹, Δ² : Ctxt) : Subst Γ Δ¹ -> Subst Γ Δ² -> Set
ExtEqˡ⁼ {Γ} ρ¹ ρ² =
  (σ : Ty Γ) -> (v : Γ ∋ σ) -> var v /⊢ ρ¹ ≈⊢ var v /⊢ ρ²

abstract

  extEqˡ⁼ (Γ : Ctxt)
          (Δ¹ : Ctxt) (ρ¹ : Subst Γ Δ¹)
          (Δ² : Ctxt) (ρ² : Subst Γ Δ²)
    : Δ¹ ≜ Δ² -> ExtEqˡ⁼ ρ¹ ρ² -> ρ¹ /≘/ ρ²
  extEqˡ⁼ {Γ} {Δ¹} {ρ¹} {Δ²} {ρ²} eqΔ proof =
    extEq (reflCtxt Γ) eqΔ
          (\σ¹ v¹ σ² v² eqv ->
              proof σ¹ v¹
                `transTm`
              substTmCong (varCong eqv) (reflSubst ρ²))

-- Another variant, useful when also the resulting contexts are
-- definitionally equal.

  extEq⁼ (Γ, Δ : Ctxt) (ρ¹ : Subst Γ Δ) (ρ² : Subst Γ Δ)
    : ExtEqˡ⁼ ρ¹ ρ² -> ρ¹ /≘/ ρ²
  extEq⁼ {Γ} {Δ} proof = extEqˡ⁼ (reflCtxt Δ) proof
