------------------------------------------------------------------------
-- Miscellaneous things
------------------------------------------------------------------------

-- # import Explicit.Types
-- # import Explicit.Equivalence
-- # import Explicit.SubstitutionLemmas
-- # import Explicit.IndicesEqual

------------------------------------------------------------------------
-- Some missing congruences and lemmas

abstract

  noVarInEmptyCtxt (whatever : Set) (σ : Ty ε)
    : ε ∋ σ -> whatever
  noVarInEmptyCtxt (v ∷≡∋ _) = noVarInEmptyCtxt v

  castTmEqˡ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
                        (τ³ : Ty Γ¹)
    : (eqτ : τ¹ ≛ τ³) -> t¹ ≈⊢ t² -> t¹ ∷≡⊢ eqτ ≈⊢ t²
  castTmEqˡ {Γ¹} {τ¹} {t¹} eqτ eqt = castTmEq t¹ eqτ `transTm` eqt

  castTmEqʳ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
                        (τ³ : Ty Γ²)
    : (eqτ : τ² ≛ τ³) -> t¹ ≈⊢ t² -> t¹ ≈⊢ t² ∷≡⊢ eqτ
  castTmEqʳ eqτ eqt = symTm (castTmEqˡ eqτ (symTm eqt))

  castTmCong (Γ¹, Γ² : Ctxt) (τ¹¹, τ¹² : Ty Γ¹) (τ²¹, τ²² : Ty Γ²)
             (t¹ : Γ¹ ⊢ τ¹¹) (t² : Γ² ⊢ τ²¹)
    :  (eqτ¹ : τ¹¹ ≛ τ¹²) -> (eqτ² : τ²¹ ≛ τ²²)
    -> t¹ ≈⊢ t² -> t¹ ∷≡⊢ eqτ¹ ≈⊢ t² ∷≡⊢ eqτ²
  castTmCong eqτ¹ eqτ² eqt = castTmEqʳ eqτ² $ castTmEqˡ eqτ¹ eqt

  appendCong (Γ¹ : Ctxt) (Γ'¹ : RCtxt Γ¹)
             (Γ² : Ctxt) (Γ'² : RCtxt Γ²)
    : Γ'¹ ≜→ Γ'² -> Γ¹ ++ Γ'¹ ≜ Γ² ++ Γ'²
  appendCong (εʳCong eqΓ)      = eqΓ
  appendCong (eqΓ' ▻ʳCong eqτ) = appendCong eqΓ' ▻Cong eqτ

------------------------------------------------------------------------
-- An improved variant of last'

last (Γ, Δ : Ctxt) (τ : Ty Γ)
  : (ρ : Subst (Γ ▻ τ) Δ) -> Δ ⊢ τ / init ρ
last {Γ} {Δ} {τ} ρ = last' ρ ∷≡⊢ symTy (compSplitsTy τ (wk τ) ρ)

abstract

  last'EqLastEq (Γ¹, Δ¹ : Ctxt) (τ¹ : Ty Γ¹) (ρ¹ : Subst (Γ¹ ▻ τ¹) Δ¹)
                (Γ², Δ² : Ctxt) (τ² : Ty Γ²) (ρ² : Subst (Γ² ▻ τ²) Δ²)
    : last' ρ¹ ≈⊢ last' ρ² -> last ρ¹ ≈⊢ last ρ²
  last'EqLastEq eqt = castTmCong _ _ eqt

------------------------------------------------------------------------
-- More missing congruences

abstract

  initCong (Γ¹, Δ¹ : Ctxt) (τ¹ : Ty Γ¹) (ρ¹ : Subst (Γ¹ ▻ τ¹) Δ¹)
           (Γ², Δ² : Ctxt) (τ² : Ty Γ²) (ρ² : Subst (Γ² ▻ τ²) Δ²)
    : ρ¹ ≘ ρ² -> init ρ¹ ≘ init ρ²
  initCong eqρ = compCong (wkCong eqτ) eqρ
    where eqτ = snocEqLastEq $ substEqCtxtEqLeft eqρ

  lastCong (Γ¹, Δ¹ : Ctxt) (τ¹ : Ty Γ¹) (ρ¹ : Subst (Γ¹ ▻ τ¹) Δ¹)
           (Γ², Δ² : Ctxt) (τ² : Ty Γ²) (ρ² : Subst (Γ² ▻ τ²) Δ²)
    : ρ¹ ≘ ρ² -> last ρ¹ ≈⊢ last ρ²
  lastCong eqρ = castTmCong _ _ $ substTmCong (varCong (vzCong eqτ)) eqρ
    where eqτ = snocEqLastEq $ substEqCtxtEqLeft eqρ

------------------------------------------------------------------------
-- Singleton right contexts

singletonʳ (Γ : Ctxt) : (σ : Ty Γ) -> RCtxt Γ
singletonʳ σ = εʳ _ ▻ʳ σ

abstract

  singletonʳCong (Γ¹ : Ctxt) (σ¹ : Ty Γ¹)
                 (Γ² : Ctxt) (σ² : Ty Γ²)
    : σ¹ ≛ σ² -> singletonʳ σ¹ ≜→ singletonʳ σ²
  singletonʳCong eqσ = εʳCong (tyEqCtxtEq eqσ) ▻ʳCong eqσ
