------------------------------------------------------------------------
-- Completeness
------------------------------------------------------------------------

-- # import Explicit
-- # import Implicit.Terms
-- # import Implicit.Substitutions
-- # import Implicit.FromExplicit
-- # import Implicit.Equality
-- # import Implicit.Equivalence
-- # import Implicit.EqualityLemmas
-- # import Implicit.CastEq
-- # import Implicit.WeakeningCong
-- # import Implicit.Coercions
-- # import Implicit.Soundness
-- # import Implicit.SubstitutionLemmas
-- # import EqKind
-- # import TmVar

------------------------------------------------------------------------
-- Equality is preserved

-- These congruences show that this view is complete with respect to
-- the explicit substitution theory.

mutual

  tmToTm⁻Cong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : t¹ ≈⊢ t² -> tmToTm⁻ t¹ ≈⊢⁻ tmToTm⁻ t²
  tmToTm⁻Cong (reflTm t)            = reflTm⁻ (tmToTm⁻ t)
  tmToTm⁻Cong (symTm eqt)           = symTm⁻ (tmToTm⁻Cong eqt)
  tmToTm⁻Cong (transTm eqt¹² eqt²³) =
    transTm⁻ (tmToTm⁻Cong eqt¹²) (tmToTm⁻Cong eqt²³)
  tmToTm⁻Cong (varCong eqv)       = var⁻Cong eqv
  tmToTm⁻Cong (λCong eqt)         = λ⁻Cong (tmToTm⁻Cong eqt)
  tmToTm⁻Cong (appCong eqt₁ eqt₂) =
    app⁻Cong (tmToTm⁻Cong eqt₁) (tmToTm⁻Cong eqt₂)
  tmToTm⁻Cong (substTmCong eqt eqρ) =
    substTm⁻Cong (tmToTm⁻Cong eqt) (substToSubst⁻Cong eqρ)
  tmToTm⁻Cong (castTmEq t eqτ)   = castTm⁻Eq (tmToTm⁻ t) _
  tmToTm⁻Cong (β t₁ t₂)          = β⁻ {Term} _ _
  tmToTm⁻Cong (η t)              = η⁻ _
  tmToTm⁻Cong (idVanishesTm t)   = idVanishesTm⁻ Term (tmToTm⁻ t)
  tmToTm⁻Cong (substWk v σ)      = subst⁻Wk Term Term v σ
  tmToTm⁻Cong (substVzSub σ t)   = subst⁻VzSub (tmToTm⁻ t)
  tmToTm⁻Cong (substVsSub v σ t) = subst⁻VsSub Term v (tmToTm⁻ t)
  tmToTm⁻Cong (substVzLift σ ρ)  = subst⁻VzLift Term σ (substToSubst⁻ ρ)
  tmToTm⁻Cong (substVsLift v σ ρ) =
    subst⁻VsLift v σ (substToSubst⁻ ρ) Term Term
  tmToTm⁻Cong (substλ t ρ) = castTm⁻Eq _ _ `transTm⁻` λ⁻Cong (reflTm⁻ _)
  tmToTm⁻Cong (substApp t₁ t₂ ρ)     = castTm⁻Eq _ _
  tmToTm⁻Cong (compSplitsTm t ρ₁ ρ₂) =
    compSplitsTm⁻ (tmToTm⁻ t) (substToSubst⁻ ρ₁) (substToSubst⁻ ρ₂)

  substToSubst⁻Cong (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
                    (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
    : ρ¹ ≘ ρ² -> substToSubst⁻ ρ¹ ≘⁻ substToSubst⁻ ρ²
  substToSubst⁻Cong (emptyCong eqΔ) =
    betweenSubst⁻ (substToSubst⁻∅ _) (substToSubst⁻∅ _) $
      ∅⁻Cong Term Term eqΔ
  substToSubst⁻Cong (snocCong eqρ eqτ eqt) =
    betweenSubst⁻ (substToSubst⁻▷ _) (substToSubst⁻▷ _) $
      Cong▷⁻ (substToSubst⁻Cong eqρ)
             eqτ
             (castTm⁻Cong _ _ $ tmToTm⁻Cong eqt)

abstract

  tyToTy⁻Cong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
              (Γ² : Ctxt) (τ² : Ty Γ²)
    : τ¹ ≛ τ² -> tyToTy⁻ τ¹ ≛⁻ tyToTy⁻ τ²
  tyToTy⁻Cong (✶Cong eqΓ)       = ✶⁻Cong eqΓ
  tyToTy⁻Cong (ΠCong eqτ₁ eqτ₂) =
    Π⁻Cong (tyToTy⁻Cong eqτ₁) (tyToTy⁻Cong eqτ₂)
  tyToTy⁻Cong (ElCong eqt) = El⁻Cong $ tmToTm⁻Cong eqt

------------------------------------------------------------------------
-- The translation functions are inverses

abstract

  tmToTm⁻LeftInv (k : EqKind) (tv : TmVar)
                 (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (t⁻ : Tm⁻ tv t) -> Tm⁻Eq k (tmToTm⁻ $ tm⁻ToTm t⁻) t⁻
  tmToTm⁻LeftInv (var⁻ _ v)    = var⁻Cong $ reflVar v
  tmToTm⁻LeftInv (λ⁻ _ t⁻)     = λ⁻Cong (tmToTm⁻LeftInv t⁻)
  tmToTm⁻LeftInv (t⁻ ∷≡⊢⁻ eqt) = castTm⁻Cong _ _ (tmToTm⁻LeftInv t⁻)
  tmToTm⁻LeftInv (t⁻₁ @⁻ t⁻₂)  =
    castTm⁻Eqˡ _ $ app⁻Cong (tmToTm⁻LeftInv t⁻₁) (tmToTm⁻LeftInv t⁻₂)

  tmToTm⁻RightInv (Γ : Ctxt) (τ : Ty Γ)
    : (t : Γ ⊢ τ) -> tm⁻ToTm (tmToTm⁻ t) ≈⊢ t
  tmToTm⁻RightInv t = tm⁻ToTmEq (tmToTm⁻ t)

  tyToTy⁻LeftInv (Γ : Ctxt) (τ : Ty Γ)
    : (τ⁻ : Ty⁻ τ) -> tyToTy⁻ (ty⁻ToTy τ⁻) ≛⁻ τ⁻
  tyToTy⁻LeftInv (✶⁻ _)        = reflTy⁻ _
  tyToTy⁻LeftInv (El⁻ t⁻)      = El⁻Cong $ tmToTm⁻LeftInv t⁻
  tyToTy⁻LeftInv (τ⁻ ∷≡✶⁻ eqτ) = castTy⁻Eqʳ _ $ tyToTy⁻LeftInv τ⁻
  tyToTy⁻LeftInv (Π⁻ τ⁻₁ τ⁻₂)  =
    Π⁻Cong (tyToTy⁻LeftInv τ⁻₁)
           (tyToTy⁻Cong (symTy (castTyEq (ty⁻ToTy τ⁻₂) _))
              `transTy⁻`
            tyToTy⁻LeftInv τ⁻₂)

  tyToTy⁻RightInv (Γ : Ctxt)
    : (τ : Ty Γ) -> ty⁻ToTy (tyToTy⁻ τ) ≛ τ
  tyToTy⁻RightInv τ = ty⁻ToTyEq (tyToTy⁻ τ)

  substToSubst⁻▷Is▷ (Γ, Δ : Ctxt) (τ : Ty Γ)
    :  (ρ : Subst Γ Δ) -> (t : Δ ⊢ τ / ρ)
    -> substToSubst⁻ (ρ ▷ t) ≘⁻ substToSubst⁻ ρ ▷⁻ tmToTm⁻ t
  substToSubst⁻▷Is▷ ρ t =
    substToSubst⁻▷ (ρ ▷ t)
      `transSubst⁻`
    Cong▷⁻ (substToSubst⁻Cong $ initIsInit _ ρ t)
           (reflTy _)
           (tmToTm⁻Cong $ lastIsLast _ ρ t)

  substToSubst⁻LeftInv (tv : TmVar) (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
    :  (ρ⁻ : Subst⁻ tv ρ)
    -> substToSubst⁻ (subst⁻ToSubst ρ⁻) ≘⁻ ρ⁻
  substToSubst⁻LeftInv (∅⁻ (Term' Var refl) Δ) = substToSubst⁻∅ (∅ Δ)
  substToSubst⁻LeftInv (ρ⁻ ▷⁻ t⁻)              =
    substToSubst⁻▷Is▷ (subst⁻ToSubst ρ⁻) (tm⁻ToTm t⁻ ∷≡⊢ _)
      `transSubst⁻`
    Cong▷⁻ (substToSubst⁻LeftInv ρ⁻)
           (reflTy _)
           (tmToTm⁻Cong (castTmEq _ _)
              `transTm⁻`
            tmToTm⁻LeftInv t⁻)
  substToSubst⁻LeftInv (ρ⁻ ∷≡⇒⁻ eqt) =
    castSubst⁻Eqʳ _ $ substToSubst⁻LeftInv ρ⁻

  substToSubst⁻RightInv (Γ, Δ : Ctxt)
    : (ρ : Subst Γ Δ) -> subst⁻ToSubst (substToSubst⁻ ρ) ≘ ρ
  substToSubst⁻RightInv ρ = subst⁻ToSubstEq (substToSubst⁻ ρ)

------------------------------------------------------------------------
-- Values are equal if they have equal indices

abstract

  tmEqTm⁻Eq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
            (tv¹ : TmVar) (t⁻¹ : Tm⁻ tv¹ t¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
            (tv² : TmVar) (t⁻² : Tm⁻ tv² t²)
    : t¹ ≈⊢ t² -> t⁻¹ ≈⊢⁻ t⁻²
  tmEqTm⁻Eq eqt =
    betweenTm⁻ (symTm⁻ $ tmToTm⁻LeftInv _) (symTm⁻ $ tmToTm⁻LeftInv _) $
      tmToTm⁻Cong $
        betweenTm (tm⁻ToTmEq _) (tm⁻ToTmEq _) eqt

  tyEqTy⁻Eq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (τ⁻¹ : Ty⁻ τ¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (τ⁻² : Ty⁻ τ²)
    : τ¹ ≛ τ² -> τ⁻¹ ≛⁻ τ⁻²
  tyEqTy⁻Eq eqτ =
    betweenTy⁻ (symTy⁻ $ tyToTy⁻LeftInv _) (symTy⁻ $ tyToTy⁻LeftInv _) $
      tyToTy⁻Cong $
        betweenTy (ty⁻ToTyEq _) (ty⁻ToTyEq _) eqτ

  substEqSubst⁻Eq (tv¹ : TmVar) (Γ¹, Δ¹ : Ctxt)
                  (ρ¹ : Subst Γ¹ Δ¹) (ρ⁻¹ : Subst⁻ tv¹ ρ¹)
                  (tv² : TmVar) (Γ², Δ² : Ctxt)
                  (ρ² : Subst Γ² Δ²) (ρ⁻² : Subst⁻ tv² ρ²)
    : ρ¹ ≘ ρ² -> ρ⁻¹ ≘⁻ ρ⁻²
  substEqSubst⁻Eq eqρ =
    betweenSubst⁻ (symSubst⁻ $ substToSubst⁻LeftInv _)
                  (symSubst⁻ $ substToSubst⁻LeftInv _) $
      substToSubst⁻Cong $
        betweenSubst (subst⁻ToSubstEq _) (subst⁻ToSubstEq _) eqρ

Implicit-Completeness : Type
Implicit-Completeness = time-plugin(Implicit-Completeness, Set)
