------------------------------------------------------------------------
-- Substitutions
------------------------------------------------------------------------

-- This view uses a set of basic substitutions differing from that in
-- Explicit. This set of substitutions is perhaps more traditional,
-- and it's not unlikely that it's easier to prove that this view is
-- equivalent to a "traditional" one. However, it's also possible that
-- the change of basic substitutions has made the completeness proof
-- more complicated. Using the same basic substitutions as in Explicit
-- might have led to other problems, though, for instance
-- termination-related ones.

-- # import Explicit
-- # import Implicit.Terms
-- # import TmVar

infixl 85 ⊙⁻
infixl 80 ↑⁻, ↑⋆⁻
infixl 70 /⁻, /∋⁻, /⊢⁻
infixl 50 ▷⁻
infixl 45 ∷≡⇒⁻, ∷⇒⁻

------------------------------------------------------------------------
-- Substitutions

mutual

  Subst⁻ (Γ, Δ : Ctxt) : TmVar -> Subst Γ Δ -> Set
  Subst⁻ {Γ} {Δ} tv ρ = Subst⁻' tv Γ Δ ρ

  data Subst⁻' : TmVar -> (Γ, Δ : Ctxt) -> Subst Γ Δ -> Set where
    ∅⁻ : (tv : TmVar) -> (Δ : Ctxt) -> Subst⁻ tv (∅ Δ)
    (▷⁻) (tv : TmVar) (Γ, Δ : Ctxt) (σ : Ty Γ)
         (ρ : Subst Γ Δ) (t : Δ ⊢ σ / ρ)
      : Subst⁻ tv ρ -> Tm⁻ tv t -> Subst⁻ tv (ρ ▷ t)
    (∷≡⇒⁻) (tv : TmVar) (Γ, Δ : Ctxt) (ρ¹, ρ² : Subst Γ Δ)
      : Subst⁻ tv ρ¹ -> ρ¹ ≘ ρ² -> Subst⁻ tv ρ²

------------------------------------------------------------------------
-- Converting Implicit substitutions to Explicit ones

mutual

  subst⁻ToSubst (tv : TmVar) (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
    : Subst⁻ tv ρ -> Subst Γ Δ
  subst⁻ToSubst (∅⁻ tv Δ) = ∅ Δ
  subst⁻ToSubst (ρ ▷⁻ t⁻) =
    subst⁻ToSubst ρ ▷
    (tm⁻ToTm t⁻
     ∷≡⊢ substTyCong (reflTy _) (symSubst $ subst⁻ToSubstEq ρ))
  subst⁻ToSubst (ρ ∷≡⇒⁻ eqρ) = subst⁻ToSubst ρ

  subst⁻ToSubstEq (tv : TmVar) (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
    : (ρ⁻ : Subst⁻ tv ρ) -> subst⁻ToSubst ρ⁻ ≘ ρ
  subst⁻ToSubstEq (∅⁻ tv Δ) = ∅Cong (reflCtxt Δ)
  subst⁻ToSubstEq (ρ ▷⁻ t⁻) =
    Cong▷ (subst⁻ToSubstEq ρ)
          (reflTy _)
          (castTmEqˡ (substTyCong (reflTy _)
                                  (symSubst $ subst⁻ToSubstEq ρ))
                     (tm⁻ToTmEq t⁻))
  subst⁻ToSubstEq (ρ ∷≡⇒⁻ eqρ) = subst⁻ToSubstEq ρ `transSubst` eqρ

------------------------------------------------------------------------
-- Substitution casts

castSubst⁻TmVar (tv¹, tv² : TmVar) (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
  : tv¹ ≡ tv² -> Subst⁻ tv¹ ρ -> Subst⁻ tv² ρ
castSubst⁻TmVar refl ρ⁻ = ρ⁻

(∷⇒⁻) (tv : TmVar)
      (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
      (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
  : Subst⁻ tv ρ¹ -> ρ¹ ≘ ρ² -> Subst⁻ tv ρ²
∅⁻ tv Δ ∷⇒⁻ emptyCong eqΔ = ∅⁻ tv _ ∷≡⇒⁻ symSubst (emptyCtxtSubstIs∅ _)
((▷⁻) {_} {_} {_} {_} {ρ¹} {t} ρ t⁻) ∷⇒⁻
  snocCong {_} {_} {_} {_} {_} {_} {_} {ρ²} eqρ eqτ eqt =
  (ρ ∷⇒⁻ lemma₁) ▷⁻ (t⁻ ∷⊢⁻ lemma₂)
  ∷≡⇒⁻ symSubst (nonEmptyCtxtSubstIs▷ ρ²)
  where
  lemma₁ = symSubst (initIsInit _ ρ¹ t) `transSubst` eqρ
  lemma₂ = symTm    (lastIsLast _ ρ¹ t) `transTm`    last'EqLastEq eqt
(ρ ∷≡⇒⁻ eqρ₁) ∷⇒⁻ eqρ₂ = ρ ∷⇒⁻ eqρ₁ `transSubst` eqρ₂

------------------------------------------------------------------------
-- Substitution application and specific substitutions

-- The implementation of substitution application borrows ideas (most
-- notably the handling of weakenings) from "Type-Preserving Renaming
-- and Substitution" (Conor McBride).

(/∋⁻) (tv : TmVar) (Γ, Δ : Ctxt) (τ : Ty Γ) (ρ : Subst Γ Δ)
  : (v : Γ ∋ τ) -> Subst⁻ tv ρ -> Tm⁻ tv (var v /⊢ ρ)
v      /∋⁻ (ρ ∷≡⇒⁻ eqρ) = v /∋⁻ ρ ∷≡⊢⁻ substTmCong (reflTm $ var v) eqρ
vz σ   /∋⁻ (ρ ▷⁻ t⁻)    = t⁻      ∷≡⊢⁻ (symTm $ substVzSub' σ _ _)
vs v σ /∋⁻ (ρ ▷⁻ t⁻)    = v /∋⁻ ρ ∷≡⊢⁻ (symTm $ substVsSub' v σ _ _)
(v ∷≡∋ eqτ) /∋⁻ ρ       = v /∋⁻ ρ ∷≡⊢⁻ (symTm $ substCastVar v eqτ _)

-- Term weakenings cannot (?) be defined right away. Instead other
-- substitutions are parameterised on Weakenings. After a while some
-- weakenings can be defined, and the parameters can be instantiated.

data Weakening (tv : TmVar) : Set where
  weakening
    :  (   (Γ : Ctxt) -> (τ : Ty Γ) -> (t : Γ ⊢ τ)
        -> Tm⁻ tv t -> (σ : Ty Γ)
        -> Tm⁻ tv (t /⊢ wk σ))
    -> Weakening tv

applyWeakening (tv : TmVar) (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
  :  Weakening tv -> Tm⁻ tv t -> (σ : Ty Γ)
  -> Tm⁻ tv (t /⊢ wk σ)
applyWeakening (weakening f) = f _ _ _

wk⇒⁻' (tv : TmVar) (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
  : Weakening tv -> Subst⁻ tv ρ -> (σ : Ty Δ) -> Subst⁻ tv (wk⇒ ρ σ)
wk⇒⁻' w (ρ ∷≡⇒⁻ eqρ) σ = wk⇒⁻' w ρ σ ∷≡⇒⁻ (wk⇒Cong eqρ (reflTy σ))
wk⇒⁻' w (∅⁻ tv Δ)    σ = ∅⁻ tv (Δ ▻ σ) ∷≡⇒⁻ symSubst (∅LeftZero $ wk σ)
wk⇒⁻' w ((▷⁻) {tv} {Γ} {Δ} {τ} {ρ} {t} ρ⁻ t⁻) σ =
  wk⇒⁻' w ρ⁻ σ ▷⁻ (applyWeakening w t⁻ σ ∷≛≡⊢⁻ eqτ)
  ∷≡⇒⁻ lemma
  where
  abstract
    eqτ : τ / ρ / wk σ ≛ τ / ρ ⊙ wk σ
    eqτ = symTy (compSplitsTy τ ρ (wk σ))

  lemma = symSubst (snocCompIsSnoc ρ (wk σ) τ t eqτ)

lift⁻' (tv : TmVar) (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
  : Weakening tv -> Subst⁻ tv ρ -> (σ : Ty Γ) -> Subst⁻ tv (ρ ↑ σ)
lift⁻' {tv} {Γ} {Δ} {ρ} w ρ⁻ σ =
  wk⇒⁻' w ρ⁻ (σ / ρ) ▷⁻ var⁻ tv (vz (σ / ρ) ∷≡∋ eqτ)
  ∷≡⇒⁻ lemma
  where
  abstract
    eqτ : σ / ρ / wk (σ / ρ) ≛ σ / wk⇒ ρ (σ / ρ)
    eqτ = symTy $ compSplitsTy σ ρ (wk (σ / ρ))

    lemma : wk⇒ ρ (σ / ρ) ▷ var (vz (σ / ρ) ∷≡∋ eqτ) ≘ ρ ↑ σ
    lemma = compWkSnocIsLift ρ σ eqτ

id⁻' (tv : TmVar)
  : Weakening tv -> (Γ : Ctxt) -> Subst⁻ tv (id Γ)
id⁻' {tv} w ε = ∅⁻ tv ε ∷≡⇒⁻ lemma
  where
  abstract
    lemma : ∅ ε ≘ id ε
    lemma = symSubst (emptyCtxtSubstIs∅ $ id ε)
id⁻' {tv} w (Γ ▻ σ) = lift⁻' w (id⁻' w Γ) σ ∷⇒⁻ lemma
  where
  abstract
    lemma : id Γ ↑ σ ≘ id (Γ ▻ σ)
    lemma = idLiftIsId σ

wk⁻' (tv : TmVar) (Γ : Ctxt)
  : Weakening tv -> (σ : Ty Γ) -> Subst⁻ tv (wk σ)
wk⁻' {tv} {Γ} w σ = wk⇒⁻' w (id⁻' w Γ) σ ∷≡⇒⁻ wk⇒IdIsWk σ

mutual

  substTy⁻' (tv : TmVar) (Γ, Δ : Ctxt) (τ : Ty Γ) (ρ : Subst Γ Δ)
    : Weakening tv -> Ty⁻ τ -> Subst⁻ tv ρ -> Ty⁻ (τ / ρ)
  substTy⁻' w (✶⁻ _)       ρ = ✶⁻ _
  substTy⁻' w (El⁻ t)      ρ = El⁻ (substTm⁻' w t ρ)
  substTy⁻' w (Π⁻ τ⁻₁ τ⁻₂) ρ =
    Π⁻ (substTy⁻' w τ⁻₁ ρ) (substTy⁻' w τ⁻₂ (lift⁻' w ρ _))
  substTy⁻' w (τ⁻ ∷≡✶⁻ eqτ) ρ =
    substTy⁻' w τ⁻ ρ
    ∷≡✶⁻ substTyCong eqτ (reflSubst _)

  substTm⁻' (tv₁, tv₂ : TmVar) (Γ, Δ : Ctxt) (τ : Ty Γ)
            (ρ : Subst Γ Δ) (t : Γ ⊢ τ)
    :  Weakening tv₂
    -> Tm⁻ tv₁ t -> Subst⁻ tv₂ ρ -> Tm⁻ (tv₁ ⊔ tv₂) (t /⊢ ρ)
  substTm⁻' _ (var⁻ tv₁ v) ρ = Tm⁻⊔ tv₁ $ v /∋⁻ ρ
  substTm⁻' w (λ⁻ τ⁻₁ t)   ρ =
    λ⁻ (substTy⁻' w τ⁻₁ ρ) (substTm⁻' w t (lift⁻' w ρ _))
    ∷≡⊢⁻ symTm (substλ _ _)
  substTm⁻' w (t₁ @⁻ t₂) ρ =
    (substTm⁻' w t₁ ρ) @⁻ (substTm⁻' w t₂ ρ)
    ∷≡⊢⁻ symTm (substApp _ _ _)
  substTm⁻' w (t ∷≡⊢⁻ eqt) ρ =
    substTm⁻' w t ρ
    ∷≡⊢⁻ substTmCong eqt (reflSubst _)

weakVar : Weakening Var
weakVar = weakening $ \_ _ _ -> helper
  where
  helper (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : Tm⁻ Var t -> (σ : Ty Γ) -> Tm⁻ Var (t /⊢ wk σ)
  helper (var⁻ Var v)  σ = var⁻ Var (vs v σ) ∷≡⊢⁻ symTm (substWk v σ)
  helper (t⁻ ∷≡⊢⁻ eqt) σ =
    helper t⁻ σ ∷≡⊢⁻ (substTmCong eqt $ reflSubst (wk σ))

weakTm : Weakening Term
weakTm = weakening $ \_ _ _ -> helper
  where
  helper (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : Tm⁻ Term t -> (σ : Ty Γ) -> Tm⁻ Term (t /⊢ wk σ)
  helper t σ = substTm⁻' weakVar t (wk⁻' weakVar σ)

weak : (tv : TmVar) -> Weakening tv
weak Var              = weakVar
weak (Term' Var refl) = weakTm

wk⇒⁻ (tv : TmVar) (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
  : Subst⁻ tv ρ -> (σ : Ty Δ) -> Subst⁻ tv (wk⇒ ρ σ)
wk⇒⁻ {tv} = wk⇒⁻' (weak tv)

(↑⁻) (tv : TmVar) (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
  : Subst⁻ tv ρ -> (σ : Ty Γ) -> Subst⁻ tv (ρ ↑ σ)
(↑⁻) {tv} = lift⁻' (weak tv)

id⁻ : (tv : TmVar) -> (Γ : Ctxt) -> Subst⁻ tv (id Γ)
id⁻ tv = id⁻' (weak tv)

wk⁻ (Γ : Ctxt)
  : (tv : TmVar) -> (σ : Ty Γ) -> Subst⁻ tv (wk σ)
wk⁻ tv = wk⁻' (weak tv)

(/⁻) (tv : TmVar) (Γ, Δ : Ctxt) (τ : Ty Γ) (ρ : Subst Γ Δ)
  : Ty⁻ τ -> Subst⁻ tv ρ -> Ty⁻ (τ / ρ)
(/⁻) {tv} = substTy⁻' (weak tv)

(/⊢⁻) (tv₂, tv₁ : TmVar) (Γ, Δ : Ctxt) (τ : Ty Γ)
      (ρ : Subst Γ Δ) (t : Γ ⊢ τ)
  : Tm⁻ tv₁ t -> Subst⁻ tv₂ ρ -> Tm⁻ (tv₁ ⊔ tv₂) (t /⊢ ρ)
(/⊢⁻) {tv₂} = substTm⁻' (weak tv₂)

sub⁻ (tv : TmVar) (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
  : Tm⁻ tv t -> Subst⁻ tv (sub t)
sub⁻ {tv} {Γ} {τ} {t} t⁻ = id⁻ tv Γ ▷⁻ (t⁻ ∷≛≡⊢⁻ eqτ) ∷≡⇒⁻ lemma
  where
  abstract
    eqτ : τ ≛ τ / id Γ
    eqτ = symTy $ idVanishesTy τ

    lemma : id Γ ▷ (t ∷≡⊢ eqτ) ≘ sub t
    lemma = idSnocIsSub t eqτ

(⊙⁻) (tv₁, tv₂ : TmVar) (Γ, Δ, X : Ctxt)
     (ρ₁ : Subst Γ Δ) (ρ₂ : Subst Δ X)
  : Subst⁻ tv₁ ρ₁ -> Subst⁻ tv₂ ρ₂ -> Subst⁻ (tv₁ ⊔ tv₂) (ρ₁ ⊙ ρ₂)
(ρ₁ ∷≡⇒⁻ eqρ) ⊙⁻ ρ₂ = ρ₁ ⊙⁻ ρ₂ ∷≡⇒⁻ compCong eqρ (reflSubst _)
∅⁻ _ Δ        ⊙⁻ ρ₂ = ∅⁻ _ _  ∷≡⇒⁻ symSubst (∅LeftZero _)
(⊙⁻) {tv₁} {tv₂} {Γ ▻ τ} {Δ} {X} {ρ₁ ▷ t} {ρ₂}
     ((▷⁻) {tv₁} {Γ} {Δ} {τ} {ρ₁} {t} ρ⁻₁ t⁻) ρ⁻₂ =
  (ρ⁻₁ ⊙⁻ ρ⁻₂) ▷⁻ (t⁻ /⊢⁻ ρ⁻₂ ∷≛≡⊢⁻ eqτ) ∷≡⇒⁻ lemma
  where
  abstract
    eqτ : τ / ρ₁ / ρ₂ ≛ τ / ρ₁ ⊙ ρ₂
    eqτ = symTy $ compSplitsTy τ ρ₁ ρ₂

    lemma : (ρ₁ ⊙ ρ₂) ▷ (t /⊢ ρ₂ ∷≡⊢ eqτ) ≘ (ρ₁ ▷ t) ⊙ ρ₂
    lemma = symSubst (snocCompIsSnoc ρ₁ ρ₂ τ t eqτ)

(↑⋆⁻) (tv : TmVar) (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
  : Subst⁻ tv ρ -> (Γ' : RCtxt Γ) -> Subst⁻ tv (ρ ↑⋆ Γ')
ρ ↑⋆⁻ εʳ Γ      = ρ
ρ ↑⋆⁻ (Γ' ▻ʳ σ) = (ρ ↑⋆⁻ Γ') ↑⁻ σ

Implicit-Substitutions : Type
Implicit-Substitutions = time-plugin(Implicit-Substitutions, Set)
