------------------------------------------------------------------------
-- Normal form equality
------------------------------------------------------------------------

-- # import Explicit
-- # import Implicit
-- # import Normalisation.NormalForms

infix 40 ≈⊨, ≈⊨′

------------------------------------------------------------------------
-- Normal and atomic form equality

mutual

  -- Atomic (neutral) form equality.

  (≈⊨′) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
        (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : Atom t¹ -> Atom t² -> Set
  (≈⊨′) = AtomEq _ _ _ _ _ _

  data AtomEq :  (Γ¹ : Ctxt) -> (τ¹ : Ty Γ¹) -> (t¹ : Γ¹ ⊢ τ¹)
              -> (Γ² : Ctxt) -> (τ² : Ty Γ²) -> (t² : Γ² ⊢ τ²)
              -> Atom t¹ -> Atom t² -> Set where

    -- Congruence.

    var⊨Cong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹)
             (Γ² : Ctxt) (τ² : Ty Γ²) (v² : Γ² ∋ τ²)
      : v¹ ≈∋ v² -> var⊨ v¹ ≈⊨′ var⊨ v²
    app⊨Cong (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹))
             (t₁¹ : Γ¹ ⊢ Π τ₁¹ τ₂¹) (t₂¹ : Γ¹ ⊢ τ₁¹)
             (at₁¹ : Atom t₁¹) (nf₂¹ : NF t₂¹)
             (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²))
             (t₁² : Γ² ⊢ Π τ₁² τ₂²) (t₂² : Γ² ⊢ τ₁²)
             (at₁² : Atom t₁²) (nf₂² : NF t₂²)
      : at₁¹ ≈⊨′ at₁² -> nf₂¹ ≈⊨ nf₂² -> at₁¹ @⊨ nf₂¹ ≈⊨′ at₁² @⊨ nf₂²

    -- Casts.

    castAtomEqˡ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (at¹ : Atom t¹)
                (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (at² : Atom t²)
                            (τ³ : Ty Γ¹) (t³ : Γ¹ ⊢ τ³)
      : (eqt : t¹ ≈⊢ t³) -> at¹ ≈⊨′ at² -> at¹ ∷≡⊨′ eqt ≈⊨′ at²
    castAtomEqʳ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (at¹ : Atom t¹)
                (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (at² : Atom t²)
                            (τ³ : Ty Γ²) (t³ : Γ² ⊢ τ³)
      : (eqt : t² ≈⊢ t³) -> at¹ ≈⊨′ at² -> at¹ ≈⊨′ at² ∷≡⊨′ eqt

  -- Normal form equality.

  (≈⊨) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
       (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : NF t¹ -> NF t² -> Set
  (≈⊨) = NFEq _ _ _ _ _ _

  data NFEq :  (Γ¹ : Ctxt) -> (τ¹ : Ty Γ¹) -> (t¹ : Γ¹ ⊢ τ¹)
            -> (Γ² : Ctxt) -> (τ² : Ty Γ²) -> (t² : Γ² ⊢ τ²)
            -> NF t¹ -> NF t² -> Set where

    -- Congruence.

    ✶AtomCong (Γ¹ : Ctxt) (t¹ : Γ¹ ⊢ ✶ Γ¹) (at¹ : Atom t¹)
              (Γ² : Ctxt) (t² : Γ² ⊢ ✶ Γ²) (at² : Atom t²)
      : at¹ ≈⊨′ at² -> ✶Atom at¹ ≈⊨ ✶Atom at²
    ElAtomCong (Γ¹ : Ctxt) (t'¹ : Γ¹ ⊢ ✶ Γ¹) (t¹ : Γ¹ ⊢ El t'¹)
               (at¹ : Atom t¹)
               (Γ² : Ctxt) (t'² : Γ² ⊢ ✶ Γ²) (t² : Γ² ⊢ El t'²)
               (at² : Atom t²)
      : at¹ ≈⊨′ at² -> ElAtom at¹ ≈⊨ ElAtom at²
    λ⊨Cong (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹))
           (t¹ : Γ¹ ▻ τ₁¹ ⊢ τ₂¹) (nf¹ : NF t¹)
           (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²))
           (t² : Γ² ▻ τ₁² ⊢ τ₂²) (nf² : NF t²)
      : nf¹ ≈⊨ nf² -> λ⊨ nf¹ ≈⊨ λ⊨ nf²

    -- Casts.

    castNFEqˡ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (nf¹ : NF t¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (nf² : NF t²)
                          (τ³ : Ty Γ¹) (t³ : Γ¹ ⊢ τ³)
      : (eqt : t¹ ≈⊢ t³) -> nf¹ ≈⊨ nf² -> nf¹ ∷≡⊨ eqt ≈⊨ nf²
    castNFEqʳ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (nf¹ : NF t¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (nf² : NF t²)
                          (τ³ : Ty Γ²) (t³ : Γ² ⊢ τ³)
      : (eqt : t² ≈⊢ t³) -> nf¹ ≈⊨ nf² -> nf¹ ≈⊨ nf² ∷≡⊨ eqt

------------------------------------------------------------------------
-- Various lemmas

abstract

  castAtomCong
    (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (t₁¹ : Γ¹ ⊢ τ₁¹) (at¹ : Atom t₁¹)
                (τ₂¹ : Ty Γ¹) (t₂¹ : Γ¹ ⊢ τ₂¹)
    (Γ² : Ctxt) (τ₁² : Ty Γ²) (t₁² : Γ² ⊢ τ₁²) (at² : Atom t₁²)
                (τ₂² : Ty Γ²) (t₂² : Γ² ⊢ τ₂²)
    :  (eqt¹ : t₁¹ ≈⊢ t₂¹) -> (eqt² : t₁² ≈⊢ t₂²)
    -> at¹ ≈⊨′ at² -> at¹ ∷≡⊨′ eqt¹ ≈⊨′ at² ∷≡⊨′ eqt²
  castAtomCong _ _ eqat = castAtomEqˡ _ (castAtomEqʳ _ eqat)

  castNFCong (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (t₁¹ : Γ¹ ⊢ τ₁¹) (nf¹ : NF t₁¹)
                         (τ₂¹ : Ty Γ¹) (t₂¹ : Γ¹ ⊢ τ₂¹)
             (Γ² : Ctxt) (τ₁² : Ty Γ²) (t₁² : Γ² ⊢ τ₁²) (nf² : NF t₁²)
                         (τ₂² : Ty Γ²) (t₂² : Γ² ⊢ τ₂²)
    :  (eqt¹ : t₁¹ ≈⊢ t₂¹) -> (eqt² : t₁² ≈⊢ t₂²)
    -> nf¹ ≈⊨ nf² -> nf¹ ∷≡⊨ eqt¹ ≈⊨ nf² ∷≡⊨ eqt²
  castNFCong _ _ eqnf = castNFEqˡ _ (castNFEqʳ _ eqnf)

------------------------------------------------------------------------
-- The relations are equivalences

mutual

  reflAtom (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (at : Atom t) -> at ≈⊨′ at
  reflAtom (var⊨ v)      = var⊨Cong $ reflVar v
  reflAtom (at₁ @⊨ nf₂)  = app⊨Cong (reflAtom at₁) (reflNF nf₂)
  reflAtom (at ∷≡⊨′ eqt) = castAtomCong eqt eqt $ reflAtom at

  reflNF (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (nf : NF t) -> nf ≈⊨ nf
  reflNF (✶Atom at)   = ✶AtomCong $ reflAtom at
  reflNF (ElAtom at)  = ElAtomCong $ reflAtom at
  reflNF (λ⊨ nf)      = λ⊨Cong $ reflNF nf
  reflNF (nf ∷≡⊨ eqt) = castNFCong eqt eqt $ reflNF nf

mutual

  symAtom (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (at¹ : Atom t¹)
          (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (at² : Atom t²)
    : at¹ ≈⊨′ at² -> at² ≈⊨′ at¹
  symAtom (var⊨Cong eqv)         = var⊨Cong (symVar eqv)
  symAtom (app⊨Cong eqat₁ eqnf₂) =
    app⊨Cong (symAtom eqat₁) (symNF eqnf₂)
  symAtom (castAtomEqˡ eqt eqat) = castAtomEqʳ eqt (symAtom eqat)
  symAtom (castAtomEqʳ eqt eqat) = castAtomEqˡ eqt (symAtom eqat)

  symNF (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (nf¹ : NF t¹)
        (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (nf² : NF t²)
    : nf¹ ≈⊨ nf² -> nf² ≈⊨ nf¹
  symNF (✶AtomCong eqat)     = ✶AtomCong (symAtom eqat)
  symNF (ElAtomCong eqat)    = ElAtomCong (symAtom eqat)
  symNF (λ⊨Cong eqnf)        = λ⊨Cong (symNF eqnf)
  symNF (castNFEqˡ eqt eqnf) = castNFEqʳ eqt (symNF eqnf)
  symNF (castNFEqʳ eqt eqnf) = castNFEqˡ eqt (symNF eqnf)

mutual

  transAtom (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (at¹ : Atom t¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (at² : Atom t²)
            (Γ³ : Ctxt) (τ³ : Ty Γ³) (t³ : Γ³ ⊢ τ³) (at³ : Atom t³)
    : at¹ ≈⊨′ at² -> at² ≈⊨′ at³ -> at¹ ≈⊨′ at³
  transAtom (castAtomEqˡ eqt eqat¹) eqat² =
    castAtomEqˡ eqt (transAtom eqat¹ eqat²)
  transAtom eqat¹ (castAtomEqʳ eqt eqat²) =
    castAtomEqʳ eqt (transAtom eqat¹ eqat²)
  transAtom (castAtomEqʳ eqt eqat₁) (castAtomEqˡ eqt eqat₂) =
    transAtom eqat₁ eqat₂
  transAtom (var⊨Cong eqv¹²) (var⊨Cong eqv²³) =
    var⊨Cong (transVar eqv¹² eqv²³)
  transAtom (app⊨Cong eqat₁¹² eqnf₂¹²) (app⊨Cong eqat₁²³ eqnf₂²³) =
    app⊨Cong (transAtom eqat₁¹² eqat₁²³) (transNF eqnf₂¹² eqnf₂²³)

  transNF (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (nf¹ : NF t¹)
          (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (nf² : NF t²)
          (Γ³ : Ctxt) (τ³ : Ty Γ³) (t³ : Γ³ ⊢ τ³) (nf³ : NF t³)
    : nf¹ ≈⊨ nf² -> nf² ≈⊨ nf³ -> nf¹ ≈⊨ nf³
  transNF (castNFEqˡ eqt eqnf¹) eqnf² =
    castNFEqˡ eqt (transNF eqnf¹ eqnf²)
  transNF eqnf¹ (castNFEqʳ eqt eqnf²) =
    castNFEqʳ eqt (transNF eqnf¹ eqnf²)
  transNF (castNFEqʳ eqt eqnf₁) (castNFEqˡ eqt eqnf₂) =
    transNF eqnf₁ eqnf₂
  transNF (✶AtomCong eqat¹²) (✶AtomCong eqat²³) =
    ✶AtomCong (transAtom eqat¹² eqat²³)
  transNF (ElAtomCong eqat¹²) (ElAtomCong eqat²³) =
    ElAtomCong (transAtom eqat¹² eqat²³)
  transNF (λ⊨Cong eqnf¹²) (λ⊨Cong eqnf²³) =
    λ⊨Cong (transNF eqnf¹² eqnf²³)

------------------------------------------------------------------------
-- More lemmas

abstract

  castAtomEq (Γ : Ctxt) (τ¹ : Ty Γ) (t¹ : Γ ⊢ τ¹)
                        (τ² : Ty Γ) (t² : Γ ⊢ τ²)
    : (at : Atom t¹) -> (eqt : t¹ ≈⊢ t²) -> at ∷≡⊨′ eqt ≈⊨′ at
  castAtomEq at eqt = castAtomEqˡ eqt (reflAtom at)

  castNFEq (Γ : Ctxt) (τ¹ : Ty Γ) (t¹ : Γ ⊢ τ¹)
                      (τ² : Ty Γ) (t² : Γ ⊢ τ²)
    : (nf : NF t¹) -> (eqt : t¹ ≈⊢ t²) -> nf ∷≡⊨ eqt ≈⊨ nf
  castNFEq nf eqt = castNFEqˡ eqt (reflNF nf)

  betweenAtom (Γ₁ : Ctxt) (τ₁ : Ty Γ₁) (t₁ : Γ₁ ⊢ τ₁) (at₁ : Atom t₁)
              (Γ₂ : Ctxt) (τ₂ : Ty Γ₂) (t₂ : Γ₂ ⊢ τ₂) (at₂ : Atom t₂)
              (Γ₃ : Ctxt) (τ₃ : Ty Γ₃) (t₃ : Γ₃ ⊢ τ₃) (at₃ : Atom t₃)
              (Γ₄ : Ctxt) (τ₄ : Ty Γ₄) (t₄ : Γ₄ ⊢ τ₄) (at₄ : Atom t₄)
    : at₁ ≈⊨′ at₂ -> at₄ ≈⊨′ at₃ -> at₂ ≈⊨′ at₃ -> at₁ ≈⊨′ at₄
  betweenAtom eq₁ eq₃ eq₂ = eq₁ `transAtom` eq₂ `transAtom` symAtom eq₃

  betweenNF (Γ₁ : Ctxt) (τ₁ : Ty Γ₁) (t₁ : Γ₁ ⊢ τ₁) (nf₁ : NF t₁)
            (Γ₂ : Ctxt) (τ₂ : Ty Γ₂) (t₂ : Γ₂ ⊢ τ₂) (nf₂ : NF t₂)
            (Γ₃ : Ctxt) (τ₃ : Ty Γ₃) (t₃ : Γ₃ ⊢ τ₃) (nf₃ : NF t₃)
            (Γ₄ : Ctxt) (τ₄ : Ty Γ₄) (t₄ : Γ₄ ⊢ τ₄) (nf₄ : NF t₄)
    : nf₁ ≈⊨ nf₂ -> nf₄ ≈⊨ nf₃ -> nf₂ ≈⊨ nf₃ -> nf₁ ≈⊨ nf₄
  betweenNF eq₁ eq₃ eq₂ = eq₁ `transNF` eq₂ `transNF` symNF eq₃

------------------------------------------------------------------------
-- Equal things have equal indices

mutual

  atomEqTmEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (at¹ : Atom t¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (at² : Atom t²)
    : at¹ ≈⊨′ at² -> t¹ ≈⊢ t²
  atomEqTmEq = helper
    where
    abstract
      helper (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (at¹ : Atom t¹)
             (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (at² : Atom t²)
        : at¹ ≈⊨′ at² -> t¹ ≈⊢ t²
      helper (var⊨Cong eqv)         = varCong eqv
      helper (app⊨Cong eqat₁ eqnf₂) =
        appCong (atomEqTmEq eqat₁) (nfEqTmEq eqnf₂)
      helper (castAtomEqˡ eqt eqat) =
        symTm eqt `transTm` atomEqTmEq eqat
      helper (castAtomEqʳ eqt eqat) = atomEqTmEq eqat `transTm` eqt

  nfEqTmEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (nf¹ : NF t¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (nf² : NF t²)
    : nf¹ ≈⊨ nf² -> t¹ ≈⊢ t²
  nfEqTmEq = helper
    where
    abstract
      helper (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (nf¹ : NF t¹)
             (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (nf² : NF t²)
        : nf¹ ≈⊨ nf² -> t¹ ≈⊢ t²
      helper (✶AtomCong eqat)      = atomEqTmEq eqat
      helper (ElAtomCong eqat)     = atomEqTmEq eqat
      helper (λ⊨Cong eqnf)         = λCong (nfEqTmEq eqnf)
      helper (castNFEqˡ eqt eqnf) =
        symTm eqt `transTm` nfEqTmEq eqnf
      helper (castNFEqʳ eqt eqnf) = nfEqTmEq eqnf `transTm` eqt

------------------------------------------------------------------------
-- Everything is a congruence

abstract

  atomToTm⁻Cong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (at¹ : Atom t¹)
                (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (at² : Atom t²)
    : at¹ ≈⊨′ at² -> atomToTm⁻ at¹ ≈⊢⁻ atomToTm⁻ at²
  atomToTm⁻Cong eqat = tmEqTm⁻Eq $ atomEqTmEq eqat

  nfToTm⁻Cong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (nf¹ : NF t¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (nf² : NF t²)
    : nf¹ ≈⊨ nf² -> nfToTm⁻ nf¹ ≈⊢⁻ nfToTm⁻ nf²
  nfToTm⁻Cong eqnf = tmEqTm⁻Eq $ nfEqTmEq eqnf

abstract

  atomToTmCong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (at¹ : Atom t¹)
               (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (at² : Atom t²)
    : at¹ ≈⊨′ at² -> atomToTm at¹ ≈⊢ atomToTm at²
  atomToTmCong eqat = tm⁻ToTmCong $ atomToTm⁻Cong eqat

  nfToTmCong (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (nf¹ : NF t¹)
             (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (nf² : NF t²)
    : nf¹ ≈⊨ nf² -> nfToTm nf¹ ≈⊢ nfToTm nf²
  nfToTmCong eqnf = tm⁻ToTmCong $ nfToTm⁻Cong eqnf

mutual

  substAtomCong (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹) (ρ⁻¹ : Subst⁻ Var ρ¹)
                (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (at¹ : Atom t¹)
                (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²) (ρ⁻² : Subst⁻ Var ρ²)
                (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (at² : Atom t²)
    :  at¹ ≈⊨′ at² -> Subst⁻Eq structural ρ⁻¹ ρ⁻²
    -> at¹ /⊨′ ρ⁻¹ ≈⊨′ at² /⊨′ ρ⁻²
  substAtomCong (var⊨Cong eqv) eqρ = castAtomCong _ _ $
    var⊨Cong $ tm⁻ToVarCongˢ $ substVar⁻Cong eqv eqρ
  substAtomCong (app⊨Cong eqat₁ eqnf₂) eqρ = castAtomCong _ _ $
    app⊨Cong (substAtomCong eqat₁ eqρ) (substNFCong eqnf₂ eqρ)
  substAtomCong (castAtomEqˡ _ eqat) eqρ = castAtomEqˡ _ $
    substAtomCong eqat eqρ
  substAtomCong (castAtomEqʳ _ eqat) eqρ = castAtomEqʳ _ $
    substAtomCong eqat eqρ

  substNFCong (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹) (ρ⁻¹ : Subst⁻ Var ρ¹)
              (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (nf¹ : NF t¹)
              (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²) (ρ⁻² : Subst⁻ Var ρ²)
              (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (nf² : NF t²)
    :  nf¹ ≈⊨ nf² -> Subst⁻Eq structural ρ⁻¹ ρ⁻²
    -> nf¹ /⊨ ρ⁻¹ ≈⊨ nf² /⊨ ρ⁻²
  substNFCong (✶AtomCong eqat) eqρ =
    ✶AtomCong $ substAtomCong eqat eqρ
  substNFCong (ElAtomCong eqat) eqρ =
    ElAtomCong $ substAtomCong eqat eqρ
  substNFCong (λ⊨Cong eqnf) eqρ = castNFCong _ _ $
    λ⊨Cong $ substNFCong eqnf (lift⁻Cong eqρ eqτ₁)
    where eqτ₁ = snocEqLastEq $ tmEqCtxtEq $ nfEqTmEq eqnf
  substNFCong (castNFEqˡ _ eqnf) eqρ = castNFEqˡ _ $
    substNFCong eqnf eqρ
  substNFCong (castNFEqʳ _ eqnf) eqρ = castNFEqʳ _ $
    substNFCong eqnf eqρ

------------------------------------------------------------------------
-- Identity substitutions vanish

mutual

  idVanishesAtom (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (at : Atom t) -> at /⊨′ id⁻ Var Γ ≈⊨′ at
  idVanishesAtom = helper
    where
    abstract
      helper (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
        : (at : Atom t) -> at /⊨′ id⁻ Var Γ ≈⊨′ at
      helper (var⊨ v) = castAtomEqˡ _ $
        var⊨Cong $ tm⁻ToVarCongˢ $ idVanishesVar⁻ v
      helper (at₁ @⊨ nf₂) = castAtomEqˡ _ $
        app⊨Cong (idVanishesAtom at₁) (idVanishesNF nf₂)
      helper (at ∷≡⊨′ eqt⁻) = castAtomCong _ _ $ idVanishesAtom at

  idVanishesNF (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (nf : NF t) -> nf /⊨ id⁻ Var Γ ≈⊨ nf
  idVanishesNF = helper
    where
    abstract
      helper (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
        : (nf : NF t) -> nf /⊨ id⁻ Var Γ ≈⊨ nf
      helper (✶Atom  at) = ✶AtomCong  $ idVanishesAtom at
      helper (ElAtom at) = ElAtomCong $ idVanishesAtom at
      helper (λ⊨ nf)     = castNFEqˡ _ $
        λ⊨Cong $ substNFCong (reflNF nf) (idLiftIsId⁻ Var _)
                   `transNF`
                 idVanishesNF nf
      helper (nf ∷≡⊨ eqt⁻) = castNFCong _ _ $ idVanishesNF nf

------------------------------------------------------------------------
-- Compositions split

mutual

  compSplitsAtom (Γ, Δ, X : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
                 (ρ₁ : Subst Γ Δ) (ρ₂ : Subst Δ X)
    :  (at : Atom t) -> (ρ⁻₁ : Subst⁻ Var ρ₁) -> (ρ⁻₂ : Subst⁻ Var ρ₂)
    -> at /⊨′ ρ⁻₁ ⊙⁻ ρ⁻₂ ≈⊨′ at /⊨′ ρ⁻₁ /⊨′ ρ⁻₂
  compSplitsAtom = helper
    where
    abstract
      helper (Γ, Δ, X : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
             (ρ₁ : Subst Γ Δ) (ρ₂ : Subst Δ X)
        :  (at : Atom t)
        -> (ρ⁻₁ : Subst⁻ Var ρ₁) -> (ρ⁻₂ : Subst⁻ Var ρ₂)
        -> at /⊨′ ρ⁻₁ ⊙⁻ ρ⁻₂ ≈⊨′ at /⊨′ ρ⁻₁ /⊨′ ρ⁻₂
      helper (var⊨ v) ρ⁻₁ ρ⁻₂ = castAtomEqʳ _ $ castAtomCong _ _ $
        var⊨Cong $ tm⁻ToVarCongˢ $
          compSplitsVar⁻ v ρ⁻₁ ρ⁻₂
            `transTm⁻`
          symTm⁻ (tm⁻ToVarSubstLemma _ _)
      helper (at₁ @⊨ nf₂) ρ⁻₁ ρ⁻₂ = castAtomEqʳ _ $ castAtomCong _ _ $
        app⊨Cong (compSplitsAtom at₁ ρ⁻₁ ρ⁻₂) (compSplitsNF nf₂ ρ⁻₁ ρ⁻₂)
      helper (at ∷≡⊨′ eqt⁻) ρ⁻₁ ρ⁻₂ = castAtomCong _ _ $
        compSplitsAtom at ρ⁻₁ ρ⁻₂

  compSplitsNF (Γ, Δ, X : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
               (ρ₁ : Subst Γ Δ) (ρ₂ : Subst Δ X)
    :  (nf : NF t) -> (ρ⁻₁ : Subst⁻ Var ρ₁) -> (ρ⁻₂ : Subst⁻ Var ρ₂)
    -> nf /⊨ ρ⁻₁ ⊙⁻ ρ⁻₂ ≈⊨ nf /⊨ ρ⁻₁ /⊨ ρ⁻₂
  compSplitsNF = helper
    where
    abstract
      helper (Γ, Δ, X : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
             (ρ₁ : Subst Γ Δ) (ρ₂ : Subst Δ X)
        :  (nf : NF t) -> (ρ⁻₁ : Subst⁻ Var ρ₁) -> (ρ⁻₂ : Subst⁻ Var ρ₂)
        -> nf /⊨ ρ⁻₁ ⊙⁻ ρ⁻₂ ≈⊨ nf /⊨ ρ⁻₁ /⊨ ρ⁻₂
      helper (✶Atom  at) ρ⁻₁ ρ⁻₂ = ✶AtomCong $ compSplitsAtom at ρ⁻₁ ρ⁻₂
      helper (ElAtom at) ρ⁻₁ ρ⁻₂ =
        ElAtomCong $ compSplitsAtom at ρ⁻₁ ρ⁻₂
      helper (λ⊨ nf) ρ⁻₁ ρ⁻₂ = castNFEqʳ _ $ castNFCong _ _ $
        λ⊨Cong $ substNFCong (reflNF nf) (compLiftIsComp⁻ ρ⁻₁ ρ⁻₂ _)
                             `transNF`
                 compSplitsNF nf (ρ⁻₁ ↑⁻ _) (ρ⁻₂ ↑⁻ _)
      helper (nf ∷≡⊨ eqt⁻) ρ⁻₁ ρ⁻₂ = castNFCong _ _ $
        compSplitsNF nf ρ⁻₁ ρ⁻₂

------------------------------------------------------------------------
-- Casts do not change the semantics of things

abstract

  generalCastAtomEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
                    (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : (at : Atom t¹) -> (eqt : t¹ ≈⊢ t²) -> at ≈⊨′ at ∷⊨′ eqt
  generalCastAtomEq at eqt = castAtomEqʳ _ $
    symAtom (idVanishesAtom at)
      `transAtom`
    symAtom (substAtomCong (reflAtom at) (generalisedId⁻IsId⁻ Var _))

  generalCastNFEq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
                  (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : (nf : NF t¹) -> (eqt : t¹ ≈⊢ t²) -> nf ≈⊨ nf ∷⊨ eqt
  generalCastNFEq nf eqt = castNFEqʳ _ $
    symNF (idVanishesNF nf)
      `transNF`
    symNF (substNFCong (reflNF nf) (generalisedId⁻IsId⁻ Var _))

Normalisation-NormalFormEquality : Type
Normalisation-NormalFormEquality =
  time-plugin(Normalisation-NormalFormEquality, Set)
