------------------------------------------------------------------------
-- Normal forms
------------------------------------------------------------------------

-- # import Explicit
-- # import Implicit
-- # import TmVar

infixl 90 @⊨
infixl 70 /⊨, /⊨′
infixl 45 ∷≡⊨, ∷≡⊨′, ∷⊨, ∷⊨′

------------------------------------------------------------------------
-- Normal and atomic forms

mutual

  -- Atomic (neutral) forms.

  Atom (Γ : Ctxt) (τ : Ty Γ) : Γ ⊢ τ -> Set
  Atom = Atom' _

  data Atom' {Γ : Ctxt} : (τ : Ty Γ) -> Γ ⊢ τ -> Set where
    var⊨ (τ : Ty Γ)
      : (v : Γ ∋ τ) -> Atom (var v)
    (@⊨) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁)) (t₁ : Γ ⊢ Π τ₁ τ₂) (t₂ : Γ ⊢ τ₁)
      : Atom t₁ -> NF t₂ -> Atom (t₁ @ t₂)
    (∷≡⊨′) (τ¹, τ² : Ty Γ) (t¹ : Γ ⊢ τ¹) (t² : Γ ⊢ τ²)
      : Atom t¹ -> t¹ ≈⊢ t² -> Atom t²

  -- Normal forms.

  NF (Γ : Ctxt) (τ : Ty Γ) : Γ ⊢ τ -> Set
  NF = NF' _

  data NF' {Γ : Ctxt} : (τ : Ty Γ) -> Γ ⊢ τ -> Set where
    ✶Atom (t : Γ ⊢ ✶ Γ)
      : Atom t -> NF t
    ElAtom (t' : Γ ⊢ ✶ Γ) (t : Γ ⊢ El t')
      : Atom t -> NF t
    λ⊨ (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁)) (t : Γ ▻ τ₁ ⊢ τ₂)
      : NF t -> NF (λ t)
    (∷≡⊨) (τ¹, τ² : Ty Γ) (t¹ : Γ ⊢ τ¹) (t² : Γ ⊢ τ²)
      : NF t¹ -> t¹ ≈⊢ t² -> NF t²

------------------------------------------------------------------------
-- Conversion to terms

mutual

  atomToTm⁻ (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : Atom t -> Tm⁻ Term t
  atomToTm⁻ (var⊨ v)      = var⁻ Term v
  atomToTm⁻ (at₁ @⊨ nf₂)  = atomToTm⁻ at₁ @⁻ nfToTm⁻ nf₂
  atomToTm⁻ (at ∷≡⊨′ eqt) = atomToTm⁻ at ∷≡⊢⁻ eqt

  nfToTm⁻ (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : NF t -> Tm⁻ Term t
  nfToTm⁻ (✶Atom  at)  = atomToTm⁻ at
  nfToTm⁻ (ElAtom at)  = atomToTm⁻ at
  nfToTm⁻ (λ⊨ nf)      = λ⁻ (tyToTy⁻ _) (nfToTm⁻ nf)
  nfToTm⁻ (nf ∷≡⊨ eqt) = nfToTm⁻ nf ∷≡⊢⁻ eqt

abstract

  atomToTm⁻Eq (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (at : Atom t) -> atomToTm⁻ at ≈⊢⁻ tmToTm⁻ t
  atomToTm⁻Eq {Γ} {τ} {t} at = tmEqTm⁻Eq (reflTm t)

  nfToTm⁻Eq (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (nf : NF t) -> nfToTm⁻ nf ≈⊢⁻ tmToTm⁻ t
  nfToTm⁻Eq {Γ} {τ} {t} nf = tmEqTm⁻Eq (reflTm t)

atomToTm (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
  : Atom t -> Γ ⊢ τ
atomToTm at = tm⁻ToTm (atomToTm⁻ at)

abstract

  atomToTmEq (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (at : Atom t) -> atomToTm at ≈⊢ t
  atomToTmEq at = tm⁻ToTmEq (atomToTm⁻ at)

nfToTm (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
  : NF t -> Γ ⊢ τ
nfToTm nf = tm⁻ToTm (nfToTm⁻ nf)

abstract

  nfToTmEq (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (nf : NF t) -> nfToTm nf ≈⊢ t
  nfToTmEq nf = tm⁻ToTmEq (nfToTm⁻ nf)

------------------------------------------------------------------------
-- Applying substitutions to atomic and normal forms

mutual

  (/⊨′) (Γ, Δ : Ctxt) (ρ : Subst Γ Δ) (τ : Ty Γ) (t : Γ ⊢ τ)
    : Atom t -> (ρ⁻ : Subst⁻ Var ρ) -> Atom (t /⊢ ρ)
  (/⊨′) {Γ} {Δ} {ρ} (var⊨ v) ρ⁻ =
    var⊨ (tm⁻ToVar (v /∋⁻ ρ⁻)) ∷≡⊨′ lemma
    where
    abstract
      lemma : var (tm⁻ToVar (v /∋⁻ ρ⁻)) ≈⊢ var v /⊢ ρ
      lemma = tm⁻ToVarEq (v /∋⁻ ρ⁻)
  (at₁ @⊨ nf₂)  /⊨′ ρ⁻ = (at₁ /⊨′ ρ⁻) @⊨ (nf₂ /⊨ ρ⁻)
                         ∷≡⊨′ symTm (substApp _ _ _)
  (at ∷≡⊨′ eqt) /⊨′ ρ⁻ = at /⊨′ ρ⁻
                         ∷≡⊨′ substTmCong eqt (reflSubst _)

  (/⊨) (Γ, Δ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ) (ρ : Subst Γ Δ)
    : NF t -> (ρ⁻ : Subst⁻ Var ρ) -> NF (t /⊢ ρ)
  ✶Atom  at    /⊨ ρ⁻ = ✶Atom  (at /⊨′ ρ⁻)
  ElAtom at    /⊨ ρ⁻ = ElAtom (at /⊨′ ρ⁻)
  λ⊨ nf        /⊨ ρ⁻ = λ⊨ (nf /⊨ ρ⁻ ↑⁻ _) ∷≡⊨ symTm (substλ _ _)
  (nf ∷≡⊨ eqt) /⊨ ρ⁻ = nf /⊨ ρ⁻ ∷≡⊨ substTmCong eqt (reflSubst _)

------------------------------------------------------------------------
-- Generalised casts

abstract

  normalFormCastLemma (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
                      (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
    : t¹ ≈⊢ t² -> (eqΓ : Γ¹ ≜ Γ²) -> t¹ /⊢ generalisedId eqΓ ≈⊢ t²
  normalFormCastLemma eqt eqΓ =
    substTmCong (reflTm _) (generalisedIdIsId eqΓ)
      `transTm`
    idVanishesTm _
      `transTm`
    eqt

(∷⊨′) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
      (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
  : Atom t¹ -> t¹ ≈⊢ t² -> Atom t²
(∷⊨′) {Γ¹} {τ¹} {t¹} {Γ²} {τ²} {t²} at eqt =
  at /⊨′ generalisedId⁻ Var eqΓ ∷≡⊨′ lemma
  where
  abstract
    eqΓ : Γ¹ ≜ Γ²
    eqΓ = tmEqCtxtEq eqt

    lemma : t¹ /⊢ generalisedId eqΓ ≈⊢ t²
    lemma = normalFormCastLemma eqt eqΓ

(∷⊨) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
     (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
  : NF t¹ -> t¹ ≈⊢ t² -> NF t²
(∷⊨) {Γ¹} {τ¹} {t¹} {Γ²} {τ²} {t²} nf eqt =
  nf /⊨ generalisedId⁻ Var eqΓ ∷≡⊨ lemma
  where
  abstract
    eqΓ : Γ¹ ≜ Γ²
    eqΓ = tmEqCtxtEq eqt

    lemma : t¹ /⊢ generalisedId eqΓ ≈⊢ t²
    lemma = normalFormCastLemma eqt eqΓ

Normalisation-NormalForms : Type
Normalisation-NormalForms = time-plugin(Normalisation-NormalForms, Set)
